"""
The solution for Part 1.
"""

def get_friendly_dict(friend_list: list[tuple[str, str]]) -> dict[str, set]:
    """Calculates the first-degree friends for everyone in a social network.

    ### Parameters:
        friend_list (list): A list of 2-tuples of representing reciprocal
        friendship links.

    ### Returns:
        friendly_dict (dict): A dictionary of sets, where each set contains the
        friends of that person.

    ### Examples:
    >>> get_friendly_dict([('kim', 'sandy'), ('alex', 'sandy'),
        ('kim', 'alex'), ('kim', 'glenn')])
    {'kim': {'glenn', 'sandy', 'alex'}, 'sandy': {'kim', 'alex'},
    'alex': {'sandy', 'kim'}, 'glenn': {'kim'}}

    >>> get_friendly_dict([('kim', 'sandy'), ('sandy', 'alex'),
        ('alex', 'glenn'), ('glenn', 'kim')])
    {'kim': {'glenn', 'sandy'}, 'sandy': {'kim', 'alex'},
    'alex': {'glenn', 'sandy'}, 'glenn': {'kim', 'alex'}}
    """

    friendly_dict = {}
    for person_a, person_b in friend_list:
        if person_a not in friendly_dict:
            friendly_dict[person_a] = set()
        if person_b not in friendly_dict:
            friendly_dict[person_b] = set()
        friendly_dict[person_a].add(person_b)
        friendly_dict[person_b].add(person_a)
    return friendly_dict
