"""
The solution for Part 4.
"""

from typing import Callable

def flatten_network(bestie_dict: dict[str, set[str]]) -> list[str]:
    """Creates a list of everyone in a social network.

    ### Parameters:
        bestie_dict (dict): A dictionary of sets of friends of each individual
        in the social network.

    ### Returns:
        individuals (list): A list of every person in a social network.
    """
    people = set()
    for person in bestie_dict:
        people.add(person)
        for bestie in bestie_dict[person]:
            people.add(bestie)

    return list(people)

def besties_coverage(individuals: list[str], bestie_dict: dict[str, set[str]],
    relationship_list: list[Callable]) -> float:
    """Calculates the proportion of connected individuals, to the total size of
    the network.

    ### Parameters:
        individuals (list): A list of individuals. Assumes every individual in
        this list is in the social network.
        bestie_dict (dict): A dictionary of sets of friends of each individual
        in the social network.
        relationship_list (list): A list of functions defining relationships in
        the social network, selected from friend_besties and
        friend_second_besties.

    ### Returns:
        coverage (float): The proportion of the total number of individuals who
        are either a member of individuals or connected via one of the
        relationships in relationship_list.

    ### Examples:
    >>> besties_coverage(['glenn'], {'kim': {'sandy', 'alex', 'glenn'},
        'sandy': {'kim', 'alex'}, 'alex': {'kim', 'sandy'}, 'glenn': {'kim'}},
        [])
    0.25

    >>> besties_coverage(['glenn'], {'kim': {'sandy', 'alex', 'glenn'},
        'sandy': {'kim', 'alex'}, 'alex': {'kim', 'sandy'}, 'glenn': {'kim'}},
        [friend_besties])
    0.5

    >>> besties_coverage(['glenn'], {'kim': {'sandy', 'alex', 'glenn'},
        'sandy': {'kim', 'alex'}, 'alex': {'kim', 'sandy'}, 'glenn': {'kim'}},
        [friend_second_besties])
    0.75

    >>> besties_coverage(['glenn'], {'kim': {'sandy', 'alex', 'glenn'},
        'sandy': {'kim', 'alex'}, 'alex': {'kim', 'sandy'}, 'glenn': {'kim'}},
        [friend_besties, friend_second_besties])
    1.0
    """

    everyone = flatten_network(bestie_dict)

    # Just return the proportion if there are no relationships
    if not relationship_list:
        return len(individuals) / len(everyone)

    # Find everyone through the relationships in the relationship list
    covered = set()
    for individual in individuals:
        covered.add(individual)
        for relationship in relationship_list:
            for person in relationship(individual, bestie_dict):
                covered.add(person)

    return len(covered) / len(everyone)
